local tablet = AceLibrary("Tablet-2.0")
local deformat = AceLibrary("Deformat-2.0")
local babbleclass = AceLibrary("Babble-Class-2.2")
local crayon = AceLibrary("Crayon-2.0")
local tourist = AceLibrary("Tourist-2.0")
local abacus = AceLibrary("Abacus-2.0")
local quixote = AceLibrary("Quixote-1.0")
local dewdrop = AceLibrary("Dewdrop-2.0")

local _G = getfenv(0)

local L = AceLibrary("AceLocale-2.2"):new("QuestsFu")

QuestsFu = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceHook-2.1", "AceConsole-2.0", "AceDB-2.0", "FuBarPlugin-2.0")

QuestsFu.version = "2.0." .. string.sub("$Revision: 27512 $", 12, -3)
QuestsFu.date = string.sub("$Date: 2007-02-07 23:45:28 -0500 (Wed, 07 Feb 2007) $", 8, 17)
QuestsFu.hasIcon = true
QuestsFu.clickableTooltip = true
QuestsFu.cannotHideText = true

local optionsTable
function QuestsFu:OnInitialize()
	self:RegisterDB("QuestsFuDB")
	self:RegisterDefaults('profile', {
		autoQuestWatch = {
			zone = false,
			subzone = true,
			gained = false,
			progress = true,
		},
		partynotify = {
			message = true,
			chat = false,
			own = false,
		},
		rememberQuestgiver = true,
		showLevelsGame = true,
		showLevelsTablet = true,
		showLevelsZone = true,
		showImpossible = true,
		showArea = true,
		showCurrentAreaOnly = false,
		showClassAnyway = true,
		showCurrentAreaDescriptionOnly = false,
		showDescription = true,
		showCompletedObjectives = false,
		showInItemTooltip = true,
		soundOnComplete = true,
		wrapQuests = true,
		showTextOpt = {
			total = true,
			current = true,
			complete = false,
			lastmessage = false,
		},
		color = {
			zone = true,
			level = true,
			title = true,
			objectives = true,
		},
		useOwnTracker = true,
		tracker_data = {
			fontSizePercent=0.7, detached=true, transparency=0, 
		},
		tracker_strata = 'HIGH',
		useOwnQuestDetail = true,
		detail_data = {
			fontSizePercent=0.9, detached=true, 
		},
		detail_strata = 'HIGH',
	})
	self:RegisterDefaults('char', {
		hidden = {},
		watchedQuests = {},
		questgivers = {},
	})
	
	optionsTable = {
		handler = QuestsFu,
		type = 'group',
		args = {
			text = {
				type = 'group', name = L["Text"], order = 99,
				desc = L["Toolbar text"],
				args = {
					current = {
						type = 'toggle', name = L["Current"],
						desc = L["Show current # of quests"],
						get = function() return self.db.profile.showTextOpt.current end,
						set = function(t) self.db.profile.showTextOpt.current = t; self:UpdateText() end,
					},
					complete = {
						type = 'toggle', name = L["Completed"],
						desc = L["Show # of quests completed"],
						get = function() return self.db.profile.showTextOpt.complete end,
						set = function(t) self.db.profile.showTextOpt.complete = t; self:UpdateText() end,
					},
					total = {
						type = 'toggle', name = L["Total"],
						desc = L["Show total possible quests"],
						get = function() return self.db.profile.showTextOpt.total end,
						set = function(t) self.db.profile.showTextOpt.total = t; self:UpdateText() end,
					},
					lastmessage = {
						type = 'toggle', name = L["Last message"],
						desc = L["Show last quest status message"],
						get = function() return self.db.profile.showTextOpt.lastmessage end,
						set = function(t) self.db.profile.showTextOpt.lastmessage = t; self:UpdateText() end,
					},
				},
			},
			levels = {
				type = 'group', name = L["Levels"], order = 99,
				desc = L["Show quest levels"],
				args = {
					game = {
						type = "toggle", name = L["In Game"],
						desc = L["Show quest levels in the quest log and NPC dialogs"],
						get = function() return self.db.profile.showLevelsGame end,
						set = "ToggleShowingLevelsGame",
					},
					tablet = {
						type = "toggle", name = L["In QuestsFu"],
						desc = L["Show quest levels in the QuestsFu tooltip"],
						get = function() return self.db.profile.showLevelsTablet end,
						set = function(t) self.db.profile.showLevelsTablet = t; self:UpdateTooltip() end,
					},
					zone = {
						type = "toggle", name = L["Zone"],
						desc = L["Show zone levels in the QuestsFu tooltip"],
						get = function() return self.db.profile.showLevelsZone end,
						set = function(t) self.db.profile.showLevelsZone = t; self:UpdateTooltip() end,
					},
				},
			},
			autoQuestWatch = {
				type = 'group', name = L["Auto Quest Watch"], order = 99,
				desc = L["Automatically add quests to the tracker when appropriate"],
				args = {
					zone = {
						type = "toggle", name = L["Zone"],
						desc = L["Add quests to the tracker when you enter their zone"],
						get = function() return self.db.profile.autoQuestWatch.zone end,
						set = function(t) self.db.profile.autoQuestWatch.zone = t end,
					},
					subzone = {
						type = "toggle", name = L["Subzone"],
						desc = L["Add quests to the tracker when you enter their subzone"],
						get = function() return self.db.profile.autoQuestWatch.subzone end,
						set = function(t) self.db.profile.autoQuestWatch.subzone = t end,
					},
					progress = {
						type = "toggle", name = L["Progress"],
						desc = L["Add quests to the tracker when you make progress on them"],
						get = function() return self.db.profile.autoQuestWatch.progress end,
						set = function(t) self.db.profile.autoQuestWatch.progress = t end,
					},
					gained = {
						type = "toggle", name = L["Gained"],
						desc = L["Add quests to the tracker when they are first gained"],
						get = function() return self.db.profile.autoQuestWatch.gained end,
						set = function(t) self.db.profile.autoQuestWatch.gained = t end,
					},
				},
			},
			partynotify = {
				type = 'group', name = L["Party notifications"], order = 99,
				desc = L["Notify when events occur"],
				args = {
					message = {
						type = "toggle", name = L["Messageframe"],
						desc = L["Show in the messageframe (like normal quest notifications)"],
						get = function() return self.db.profile.partynotify.message end,
						set = function(t) self.db.profile.partynotify.message = t end,
					},
					chat = {
						type = "toggle", name = L["Chat"],
						desc = L["Show in the chatframe"],
						get = function() return self.db.profile.partynotify.chat end,
						set = function(t) self.db.profile.partynotify.chat = t end,
					},
					own = {
						type = "toggle", name = L["Announce own objectives"],
						desc = L["Announce own quest objectives in party chat, because sometimes people are crazy enough as to not be using QuestsFu."],
						get = function() return self.db.profile.partynotify.own end,
						set = function(t) self.db.profile.partynotify.own = t end,
					},
				},
			},
			impossible = {
				type = 'toggle', name = L["Impossible quests"],
				desc = L["Show impossible (red) quests"],
				get = function() return self.db.profile.showImpossible end,
				set = function(t) self.db.profile.showImpossible = t; self:UpdateTooltip() end,
			},
			area = {
				type = 'toggle', name = L["Zone"],
				desc = L["Show quest zones"],
				get = function() return self.db.profile.showArea end,
				set = function(t) self.db.profile.showArea = t; self:UpdateTooltip() end,
			},
			caonly = {
				type = 'toggle', name = L["Current area only"],
				desc = L["Only show quests for the current zone"],
				get = function() return self.db.profile.showCurrentAreaOnly end,
				set = function(t) self.db.profile.showCurrentAreaOnly = t; self:UpdateTooltip() end,
			},
			classanyway = {
				type = 'toggle', name = L["Class quests always"],
				desc = L["Always show class quests"],
				get = function() return self.db.profile.showClassAlways end,
				set = function(t) self.db.profile.showClassAlways = t; self:UpdateTooltip() end,
			},
			description = {
				type = 'toggle', name = L["Details"],
				desc = L["Show quest details"],
				get = function() return self.db.profile.showDescription end,
				set = function(t) self.db.profile.showDescription = t; self:UpdateTooltip() end,
			},
			cadonly = {
				type = 'toggle', name = L["Current area details only"],
				desc = L["Show details for current area quests only"],
				get = function() return self.db.profile.showCurrentAreaDescriptionOnly end,
				set = function(t) self.db.profile.showCurrentAreaDescriptionOnly = t; self:UpdateTooltip() end,
			},
			completed= {
				type = 'toggle', name = L["Completed objectives"],
				desc = L["Show completed objectives"],
				get = function() return self.db.profile.showCompletedObjectives end,
				set = function(t) self.db.profile.showCompletedObjectives = t; self:UpdateTooltip() end,
			},
			tooltip = {
				type = 'toggle', name = L["Objectives in tooltip"],
				desc = L["Show quest objectives in item/mob tooltips"],
				get = function() return self.db.profile.showInItemTooltip end,
				set = function(t) self.db.profile.showInItemTooltip = t; self:UpdateTooltip() end,
			},
			soundOnComplete = {
				type = 'toggle', name = L["Sound on complete"],
				desc = L["Play a sound when all a quests' objectives are complete"],
				get = function() return self.db.profile.soundOnComplete end,
				set = function(t) self.db.profile.soundOnComplete = t end,
			},
			color = {
				type = 'group', name = L["Color"], order = 99,
				desc = L["Color settings"],
				args = {
					zone = {
						type = 'toggle', name = L["Zone"],
						desc = L["Color zone names by suggested level"],
						get = function() return self.db.profile.color.zone end,
						set = function(t) self.db.profile.color.zone = t; self:UpdateTooltip() end,
					},
					title = {
						type = 'toggle', name = L["Title"],
						desc = L["Color the quest title by difficulty"],
						get = function() return self.db.profile.color.title end,
						set = function(t) self.db.profile.color.title = t; self:UpdateTooltip() end,
					},
					level = {
						type = 'toggle', name = L["Level"],
						desc = L["Color the quest level by difficulty"],
						get = function() return self.db.profile.color.level end,
						set = function(t) self.db.profile.color.level = t; self:UpdateTooltip() end,
					},
					objectives = {
						type = 'toggle', name = L["Objectives"],
						desc = L["Color objective completion status"],
						get = function() return self.db.profile.color.objectives end,
						set = function(t) self.db.profile.color.objectives = t; self:UpdateTooltip() end,
					},
				},
			},
			wrap = {
				type = 'toggle', name = L["Wrap quest titles"],
				desc = L["Wrap long quest titles on to multiple lines"],
				get = function() return self.db.profile.wrapQuests end,
				set = function(t) self.db.profile.wrapQuests = t; self:UpdateTooltip() end,
			},
			tracker = {
				type = 'group', name = L["Tracker"], handler = QuestsFu,  --Handler here because we're going to reuse this table later on in the tablet registration.
				desc = L["Options for the tracker"], order = 99,
				args = {
					enabled = {
						type = 'toggle', name = L["Use own tracker"],
						desc = L["Replace the default quest tracker with a slightly more featureful one."],
						get = function() return self.db.profile.useOwnTracker end,
						set = function(t)
							self.db.profile.useOwnTracker = t
							if self.db.profile.useOwnTracker then
								if tablet:IsRegistered('QuestsFu_Tracker') then tablet:Open('QuestsFu_Tracker') end
								self:UpdateTracker()
							else tablet:Close('QuestsFu_Tracker'); QuestWatchFrame:Show() end
						end,
					},
					strata = {
						name=L["Strata"], type="text", desc=L["Adjust the strata of this panel (takes effect after you reload your UI)"],
						get=function() return self.db.profile.tracker_strata end,
						set=function(t) self.db.profile.tracker_strata = t end,
						validate={'BACKGROUND', 'LOW', 'MEDIUM', 'HIGH', },
					},
					--This one will look a little weird on the tablet right-click menu, as it'll be duplicating native functionality.  But it's totally worth it to stop people complaining.
					lock = {
						type = 'toggle', name = L["Lock"],
						desc = L["Lock or unlock this panel"],
						disabled = function() return not tablet:IsRegistered("QuestsFu_Tracker") end,
						get = function() return tablet:IsRegistered("QuestsFu_Tracker") and tablet:IsLocked("QuestsFu_Tracker") end,
						set = function() tablet:ToggleLocked("QuestsFu_Tracker") end,
					},
				},
			},
			detail = {
				type = 'group', name = L["Quest details"], handler = QuestsFu,  --Handler here because we're going to reuse this table later on in the tablet registration.
				desc = L["Options for the details pane"], order = 99,
				args = {
					enabled = {
						type = 'toggle', name = L["Use own quest detail pane"],
						desc = L["Display detailed quest information in a custom window."],
						get = function() return self.db.profile.useOwnQuestDetail end,
						set = function(t) self.db.profile.useOwnQuestDetail = t end,
					},
					strata = {
						name=L["Strata"], type="text", desc=L["Adjust the strata of this panel (takes effect after you reload your UI)"],
						get=function() return self.db.profile.detail_strata end,
						set=function(t) self.db.profile.detail_strata = t end,
						validate={'BACKGROUND', 'LOW', 'MEDIUM', 'HIGH', },
					},
					--This one will look a little weird on the tablet right-click menu, as it'll be duplicating native functionality.  But it's totally worth it to stop people complaining.
					lock = {
						type = 'toggle', name = L["Lock"],
						desc = L["Lock or unlock this panel"],
						disabled = function() return not tablet:IsRegistered("QuestsFu_Detail") end,
						get = function() return tablet:IsLocked("QuestsFu_Detail") end,
						set = function() tablet:ToggleLocked("QuestsFu_Detail") end,
					},
				},
			},
			remember = {
				type = 'toggle', name = L["Remember Questgiver"],
				desc = L["Record who initiated a quest."],
				get = function() return self.db.profile.rememberQuestgiver end,
				set = function(t) self.db.profile.rememberQuestgiver = t end,
			}
		}
	}
	self:RegisterChatCommand({"/questsfu", "/fuq" }, optionsTable)
	self.OnMenuRequest = optionsTable
	
	self.lastuimessage = ""
	self.lastquestmessage = ""
	
	self.loadedWatchedQuests = false
	
	for questName in pairs(self.db.char.watchedQuests) do
		if type(questName) == 'number' then
			self.db.char.watchedQuests[questName] = nil
		end
	end
end

function QuestsFu:OnEnable()
	self:RegisterEvent("Quixote_Update", "Update")
	self:RegisterEvent("PLAYER_LEVEL_UP", "UpdateTooltip") -- Quest difficulty colors can change on level; just redraw the tooltip in case they have it open while levelling.
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "UpdateTooltip")
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT", "OnMouseOverUnit")
	self:RegisterEvent("UI_INFO_MESSAGE","OnUIInfoMessage")
	self:SecureHook("AddQuestWatch")
	self:SecureHook("RemoveQuestWatch")
	self:Hook("AcceptQuest", true)
	self:RegisterEvent("Quixote_Quest_Lost", "Quest_Lost")

	self:SecureHook("QuestWatch_Update", "UpdateTracker")
	self:RegisterEvent("Quixote_Party_Quest_Sync", "UpdateTracker")
	self:RegisterEvent("Quixote_Party_Quest_Gained", "UpdateTracker")
	self:RegisterEvent("Quixote_Party_Quest_Lost", "UpdateTracker")
	self:RegisterEvent("Quixote_Party_Leaderboard_Update")
	--self:Hook("GetLFGTypeEntries", true)

	self:HookScript(GameTooltip, "OnTooltipSetItem")

	self:ScheduleEvent(self.LoadWatchedQuests, 1, self)

	if self.db.profile.showLevelsGame then
		self:RegisterEvent("GOSSIP_SHOW", "OnGossipShow")
		self:RegisterEvent("QUEST_GREETING", "OnQuestGreeting")
		self:SecureHook("QuestLog_Update")
	end
	
	--AutoQuestWatch bits:
	self:RegisterEvent("MINIMAP_ZONE_CHANGED", "ZoneChange")
	self:RegisterEvent("ZONE_CHANGED", "ZoneChange")
	self:RegisterEvent("ZONE_CHANGED_INDOORS", "ZoneChange")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "ZoneChange")
	self:RegisterEvent("Quixote_Leaderboard_Update")
	self:RegisterEvent("Quixote_Quest_Gained")
	
	--For sound!
	self:RegisterEvent("Quixote_Quest_Complete")
	
	local function checkQuestgivers()
		-- Make sure that we haven't lost a quest whose questgiver we're remembering.
		for name,text in pairs(self.db.char.questgivers) do
			if not quixote:GetQuest(name) then
				self.db.char.questgivers[name] = nil
			end
		end
	end
	
	if quixote.firstScanDone then
		checkQuestgivers()
	else
		self:RegisterEvent("Quixote_Ready", checkQuestgivers, true)
	end
end

function QuestsFu:ToggleShowingLevelsGame(t)
	self.db.profile.showLevelsGame = t
	if t then
		self:RegisterEvent("GOSSIP_SHOW", "OnGossipShow")
		self:RegisterEvent("QUEST_GREETING", "OnQuestGreeting")
		self:SecureHook("QuestLog_Update")
	else
		self:UnregisterEvent("GOSSIP_SHOW", "OnGossipShow")
		self:UnregisterEvent("QUEST_GREETING", "OnQuestGreeting")
		self:Unhook("QuestLog_Update")
	end
	--self:UpdateData()
end

----------------------------------------------------
-- Add quest levels to the quest log
----------------------------------------------------

function QuestsFu:QuestLog_Update()
	--This is fragile, but lets us ignore event sequencing.  It'll break if anyone is meddling with the output of GetQuestLogTitle.
	--(questIndexes can change on every QUEST_LOG_UPDATE; this *will* be called before Quixote updates from that event, because Quixote is
	--throttling to avoid crashing the client on edge cases. I blame Blizzard for it being possible to crash the game by scanning the quest log. So.)
	for i = 1, QUESTS_DISPLAYED, 1 do
		--local questIndex = i + FauxScrollFrame_GetOffset(QuestLogListScrollFrame)
		local titleLine = _G["QuestLogTitle"..i]
		if titleLine and titleLine:GetText() then
			local title, level, tag = quixote:GetQuest(titleLine:GetText():trim())
			if title then
				titleLine:SetText(self:MakeTag(level, tag)..title)
				local check = _G["QuestLogTitle" .. i .. "Check"]
				if check:IsVisible() then
					check:SetPoint("LEFT", titleLine, "LEFT", QuestLogDummyText:GetWidth() + 24, 0)
				end
			end
		end
	end
end

----------------------------------------------------
-- Add quest levels to LFG
----------------------------------------------------

--[[
-- Fuck you, Blizzard. ~vhaarr
function QuestsFu:GetLFGTypeEntries(index)
	if index and tonumber(index) then
		index = tonumber(index)
		local original = {self.hooks["GetLFGTypeEntries"](index)}
		for i, v in ipairs(original) do
			if index == 2 or index == 3 or index == 5 then
				if (index == 5 and tourist:IsZone(v)) or tourist:IsInstance(v) then
					local low, high = tourist:GetLevel(v)
					if low and high then
						original[i] = v .. " ("..tostring(low).."-"..tostring(high)..")"
					end
				end
			elseif index == 4 then
				local _, level = quixote:GetQuestByName(v)
				if level and tonumber(level) then
					original[i] = "["..tostring(level).."] "..v
				end
			end
		end
		return unpack(original)
	end
	return self.hooks["GetLFGTypeEntries"](index)
end
]]

----------------------------------------------------
-- Add quest levels to NPC dialog when possible
----------------------------------------------------

--This is pretty much how Minimalist does it, which it credits to AutoSelect.
function QuestsFu:OnGossipShow()
	local buttonindex = 1
	local list, button
	if GetGossipAvailableQuests() then
		list = {GetGossipAvailableQuests()}
		for i = 2,getn(list),2 do
			button = _G["GossipTitleButton"..buttonindex]
			button:SetText(format('[%d] %s',list[i],list[i-1]))
			buttonindex = buttonindex + 1
		end
		buttonindex = buttonindex + 1
		for k in pairs(list) do
			list[k] = nil
		end
	end
	if GetGossipActiveQuests() then
		list = {GetGossipActiveQuests()}
		for i = 2,getn(list),2 do
			button = _G["GossipTitleButton"..buttonindex]
			button:SetText(format('[%d] %s',list[i],list[i-1]))
			buttonindex = buttonindex + 1
		end
		for k in pairs(list) do
			list[k] = nil
		end
	end
end

--This is pretty much how Minimalist does it, which it credits to AutoSelect.
function QuestsFu:OnQuestGreeting()
	local numactive,numavailable = GetNumActiveQuests(), GetNumAvailableQuests()
	local title, level, button
	local o,GetTitle,GetLevel = 0,GetActiveTitle,GetActiveLevel
	for i=1, numactive + numavailable do
		if i == numactive + 1 then
			o,GetTitle,GetLevel = numactive,GetAvailableTitle,GetAvailableLevel
		end
		title,level = GetTitle(i-o), GetLevel(i-o)
		button = _G["QuestTitleButton"..i]
		button:SetText(format('[%d] %s',level,title))
	end
end

----------------------------------------------------
-- Quest info messages
----------------------------------------------------

function QuestsFu:OnUIInfoMessage()
	self.lastuimessage = arg1
	self:ScheduleEvent(self.ClearUIInfoMessage, 0.5, self)
end

function QuestsFu:ClearUIInfoMessage()
	self.lastuimessage = ""
end

----------------------------------------------------
-- Tooltips
----------------------------------------------------

function QuestsFu:OnMouseOverUnit()
	-- Make sure that this is wanted, and that the tooltip is actually read/writeable.
	if self.db.profile.showInItemTooltip then
		--Comes with color codes.  Strip 'em out.
		if quixote:IsQuestMob(UnitName('mouseover')) then
			for _,q,i,n in quixote:IterateQuestsForMob(UnitName('mouseover')) do
				GameTooltip:AddDoubleLine(q, i..'/'..n, 1, 1, 1, crayon:GetThresholdColor(i/n))
			end
			GameTooltip:Show()
		end
	end
end

function QuestsFu:OnTooltipSetItem(tooltip, ...)
	if self.db.profile.showInItemTooltip then
		local q,i,n = quixote:IsQuestItem(tooltip:GetItem())
		if q then
			tooltip:AddDoubleLine(q, i..'/'..n, 1, 1, 1, crayon:GetThresholdColor(i/n))
			tooltip:Show()
		end
	end
	return self.hooks[tooltip].OnTooltipSetItem(tooltip, ...)
end

----------------------------------------------------
-- FuBarPlugin OnFooUpdates
----------------------------------------------------

function QuestsFu:OnDataUpdate()
	self.lastquestmessage = self.lastuimessage
	self:SaveWatchedQuests()
	--A watched quest may have gone away because of the update.  (The game doesn't call RemoveQuestWatch for it...)
	self:UpdateQuestDetail()
end

function QuestsFu:OnTextUpdate()
	local text = ""
	local maxQuests = MAX_QUESTS
	local r,g,b = GameFontNormal:GetTextColor()
	local color = string.format("%02X%02X%02X", r * 255, g * 255, b * 255)
	local numQuests, numQuestsDone = quixote:GetNumQuests()
	if self.db.profile.showTextOpt.complete then
		color = crayon:GetThresholdHexColor(numQuestsDone / numQuests)
		text = text .. crayon:Colorize(color, numQuestsDone)
	end
	if self.db.profile.showTextOpt.complete and (self.db.profile.showTextOpt.current or self.db.profile.showTextOpt.total) then
		text = text .. crayon:Colorize(color, "/")
	end
	if self.db.profile.showTextOpt.current then
		color = crayon:GetThresholdHexColor((maxQuests - numQuests) / maxQuests)
		text = text .. crayon:Colorize(color, numQuests)
	end
	if self.db.profile.showTextOpt.current and self.db.profile.showTextOpt.total then
		text = text .. crayon:Colorize(color, "/")
	end
	if self.db.profile.showTextOpt.total then
		text = text .. crayon:Colorize(color, maxQuests)
	end
	if (self.db.profile.showTextOpt.total or self.db.profile.showTextOpt.complete or self.db.profile.showTextOpt.current) and self.db.profile.showTextOpt.lastmessage and self.lastquestmessage ~= "" then
		text = text .. " m: "
	end
	if self.db.profile.showTextOpt.lastmessage and self.lastquestmessage ~= "" then
		text = text .. self.lastquestmessage
	end
	self:SetText(text)
end

function QuestsFu:OnTooltipUpdate()
	if quixote:GetNumQuests() > 0 then
		if not self.db.profile.showArea then
			cat = tablet:AddCategory('columns', 2, 'child_func', 'QuestClick', 'child_arg1', self)
			for _,questid in quixote:QuestsByLevel() do
				self:AddQuestToCategory(cat, questid, true, self.db.profile.showDescription)
			end
		else
			local currentZone = GetZoneText()
			local zonename
			local cat
			for _, zone in quixote:IterZones() do
				-- If showing area headers and either not showing only the current area's quests or this is the current area.
				if self.db.profile.showArea and ((currentZone == zone) or not self.db.profile.showCurrentAreaOnly or (self.db.profile.showClassAnyway and UnitClass("player") == zone)) then
					local r,g,b = 0.749,1,0.749
					zonename = zone
					if zone == UnitClass("player") then 
						r, g, b = babbleclass:GetColor(zone)
					else
						if self.db.profile.showLevelsZone and tourist:IsZoneOrInstance(zone) then
							local high, low = tourist:GetLevel(zone)
							if high > 0 and low > 0 then
								zonename = string.format("%s (%d-%d)", zone, high, low)
							end
						end
						if self.db.profile.color.zone and tourist:IsZoneOrInstance(zone) then
							r, g, b = tourist:GetLevelColor(zone)
						end
					end
					cat = tablet:AddCategory(
						'id', zone, 'text', zonename,
						'columns', 2,
						'hideBlankLine', true, 'showWithoutChildren', true,
						'checked', true, 'hasCheck', true, 'checkIcon', self.db.char.hidden[zone] and "Interface\\Buttons\\UI-PlusButton-Up" or "Interface\\Buttons\\UI-MinusButton-Up",
						'func', 'ToggleCategory', 'arg1', self, 'arg2', zone,
						'textR', r, 'textG', g, 'textB', b,
						'child_func', 'QuestClick', 'child_arg1', self
					)
					-- If we're not hiding the zone and it's either the current zone or we're not showing only the current zone.
					if (not self.db.char.hidden[zone]) and ((not self.db.profile.showCurrentAreaOnly) or (currentZone == zone) or (self.db.profile.showClassAnyway and UnitClass("player") == zone)) then
						for _,questid in quixote:QuestsByZone(zone) do
							self:AddQuestToCategory(cat, questid, false, self.db.profile.showDescription)
						end
					end
				end
			end
		end
	end
	tablet:SetHint(L["TOOLTIP_HINT"])
end

----------------------------------------------------
-- Tablet utility functions
----------------------------------------------------

function QuestsFu:AddQuestToCategory(cat, questref, appendZone, showObjectives, hideChecks)
	local title, level, tag, suggestedGroup, complete, leaderboard, zone, questid = quixote:GetQuest(questref)
	if not title or not level then return end
	
	--Are we hiding impossible quests, and is this quest impossible?
	local questImpossible = (GetDifficultyColor(level) == QuestDifficultyColor.impossible)
	if not questImpossible or (questImpossible and self.db.profile.showImpossible) then
		local thisQuest, thisQuestColor
		local questLevel, questZone = "",""
		local currentZone = GetZoneText()
		
		local questShared = QuestsFu:MembersOnQuest(questid)
		if questShared == 0 then questShared = "" end
		if appendZone then questZone = string.format(" (%s)", zone) end
		if self.db.profile.showLevelsTablet then questLevel = self:MakeTag(level, tag) end
		thisQuestColor = GetDifficultyColor(level)
		if complete == 1 then complete = L["(done)"]
		elseif complete == -1 then complete = L["(failed)"]
		else complete = nil end
		
		if not self.db.profile.color.title then title = crayon:White(title) end
		if not self.db.profile.color.level then questLevel = crayon:White(questLevel) end
		
		cat:AddLine(
			'text', questLevel..title..questZone..((questShared ~= "") and " |c0000ff00<"..questShared..">|r" or ''), 'wrap', self.db.profile.wrapQuests,
			'text2', complete,
			'textR', (thisQuestColor and thisQuestColor.r) or 1, 'textG', (thisQuestColor and thisQuestColor.g) or 1, 'textB', (thisQuestColor and thisQuestColor.b) or 1,
			'text2R', (complete == L["(failed)"]) and 1 or 0, 'text2G', (complete == L["(done)"]) and 1 or 0, 'text2B', 0,
			'func', 'QuestClick', 'arg1', self, 'arg2', questid,
			'indentation', 6,
			'checked', self.db.char.watchedQuests[title],
			'hasCheck', (not hideChecks) and self.db.char.watchedQuests[title] or false
		)

		-- If we're showing descriptions/objectives and it's either the current zone or we're not showing descriptions only for the current zone.
		if showObjectives and ((currentZone == zone) or not self.db.profile.showCurrentAreaDescriptionOnly) then
			-- If we know of leaderboard objectives:
			if leaderboard > 0 then
				for i=1,leaderboard do
					self:AddLeaderboardToCategory(cat, questid, i)
				end
			-- Otherwise, if the quest is incomplete (or complete and we're showing completed objectives), add the generic objective.
			elseif not (complete and not self.db.profile.showCompletedObjectives) then
				objective = quixote:GetQuestText(questid)
				cat:AddLine(
					'text', objective, 'wrap', true,
					'textR', (thisQuestColor and thisQuestColor.r) or 1, 'textG', (thisQuestColor and thisQuestColor.g) or 1, 'textB', (thisQuestColor and thisQuestColor.b) or 1,
					'size', tablet:GetNormalFontSize()-2,
					'arg2', questid,
					'indentation', 12
				)
			end
		end
	end
end

function QuestsFu:AddLeaderboardToCategory(cat, questid, i, fontSize, indent)
	local title = quixote:GetQuest(questid)
	local description, qtype, numGot, numNeeded, done = quixote:GetQuestStatusById(questid, i)
	if not fontSize then fontSize = tablet:GetNormalFontSize()-2 end
	if not indent then indent = 12 end
	if not (done and not self.db.profile.showCompletedObjectives) then
		local r,g,b
		local thisQuestColor = GetDifficultyColor(level)
		if self.db.profile.color.objectives then
			if numGot == nil then self:Print("The author of QuestsFu would very much like to know if you see this message.  Email the rest of this line to him at kemayo at gmail dot com.", quixote:GetQuestStatusById(questid, i)) end
			r,g,b = crayon:GetThresholdColor((qtype == 'reputation' and quixote:GetReactionLevel(numGot) or numGot) / (qtype == 'reputation' and quixote:GetReactionLevel(numNeeded) or numNeeded))
		end
		local party = ''
		for i=1, GetNumPartyMembers(), 1 do
			local unit = 'party'..i
			local n = quixote:GetPartyQuestStatus(unit, title, description)
			if n and type(n) == 'number' then
				if n == numNeeded then n = 'd' end
				party = string.format("%s|c00%s%s:%s|r ", party, babbleclass:GetHexColor(UnitClass(unit)), string.sub(UnitName(unit), 1, 3), n)
			end
		end
		cat:AddLine(
			'text', '  '..description,
			'text2', party .. (done and L["(done)"] or string.format("%s/%s", numGot, numNeeded)),
			'textR', r or ((thisQuestColor and thisQuestColor.r) or 1), 'textG', g or ((thisQuestColor and thisQuestColor.g) or 1), 'textB', b or ((thisQuestColor and thisQuestColor.b) or 1),
			'text2R', r or ((thisQuestColor and thisQuestColor.r) or 1), 'text2G', g or ((thisQuestColor and thisQuestColor.g) or 1), 'text2B', b or ((thisQuestColor and thisQuestColor.b) or 1),
			'size', fontSize, 'size2', fontSize,
			'arg2', questid,
			'indentation', indent
		)
	end
end

function QuestsFu:ToggleCategory(id, button)
	if self.db.char.hidden[id] then
		self.db.char.hidden[id] = false
	else
		self.db.char.hidden[id] = true
	end
	-- Refresh in place
	self:UpdateTooltip()
end

--TheFly contributed the base of this, and it's still using much of the logic he provided
function QuestsFu:QuestClick(questid)
	if IsAltKeyDown() then
		-- Shift-click toggles quest-watch on this quest.
		if self.db.profile.useOwnTracker then
			local questName = quixote:GetQuestById(questid)
			if self.db.char.watchedQuests[questName] then
				self.db.char.watchedQuests[questName] = nil
			else
				self.db.char.watchedQuests[questName] = true
			end
			self:UpdateTracker()
		else
			if IsQuestWatched(questid) then
				RemoveQuestWatch(questid)
			else
				-- Set error if no objectives
				if GetNumQuestLeaderBoards(questid) == 0 then
					UIErrorsFrame:AddMessage(QUEST_WATCH_NO_OBJECTIVES, 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME)
					return
				end
				-- Set an error message if trying to show too many quests
				if GetNumQuestWatches() >= MAX_WATCHABLE_QUESTS then
					UIErrorsFrame:AddMessage(format(QUEST_WATCH_TOO_MANY, MAX_WATCHABLE_QUESTS), 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME)
					return
				end
				AddQuestWatch(questid)
			end
			QuestWatch_Update()
		end
	elseif IsShiftKeyDown() and IsControlKeyDown() and ChatFrameEditBox:IsVisible() then
		-- Add the quest objectives to the chat editbox, if it's open.
		_,_,_,_,_,leaderboard = quixote:GetQuestById(questid)
		local i = 1
		for i=1,leaderboard do
			local description, qtype, numGot, numNeeded = quixote:GetQuestStatusById(questid, i)
			ChatFrameEditBox:Insert(string.format("{%s %s/%s} ", description, numGot, numNeeded))
		end
	elseif IsShiftKeyDown() and ChatFrameEditBox:IsVisible() then
		-- Add quest title to the chat editbox if it's open.
		local name,level,tag = quixote:GetQuestById(questid)
		ChatFrameEditBox:Insert(self:MakeTag(level, tag)..name)
	elseif IsControlKeyDown() then
		self:ShareQuest(questid)
	else
		self:ShowLog(questid)
	end
end

function QuestsFu:OnClick()
	ToggleQuestLog()
end

----------------------------------------------------
-- Quest watching
----------------------------------------------------

function QuestsFu:LoadWatchedQuests()
	if self.db.profile.useOwnTracker then
		self:UpdateTracker()
	else
		if not (GetNumQuestWatches() > 0) then
			for questName in pairs(self.db.char.watchedQuests) do
				local questId = quixote:GetQuestByName(questName)
				if questId then
					AddQuestWatch(questId)
				else
					self.db.char.watchedQuests[questName] = nil
				end
			end
		end
		QuestWatch_Update()
	end
	self.loadedWatchedQuests = true
end

function QuestsFu:SaveWatchedQuests()
	if self.db.profile.useOwnTracker then
		self:UpdateTracker()
	else
		if self.loadedWatchedQuests then
			for k in pairs(self.db.char.watchedQuests) do
				self.db.char.watchedQuests[k] = nil
			end
			for i = 1, GetNumQuestWatches() do
				local questName = quixote:GetQuestById(GetQuestIndexForWatch(i))
				if questName then self.db.char.watchedQuests[questName] = true end
			end
		end
	end
end

function QuestsFu:AddQuestWatch(questid)
	if self.db.profile.useOwnTracker then
		local questName = quixote:GetQuestById(questid)
		self.db.char.watchedQuests[questName] = true
	end
	self:SaveWatchedQuests()
	QuestWatchFrame:Hide()
end

function QuestsFu:RemoveQuestWatch(questid)
	if self.db.profile.useOwnTracker then
		local questName = quixote:GetQuestById(questid)
		if questName then
			self.db.char.watchedQuests[questName] = nil
		end
	end
	self:SaveWatchedQuests()
end

function QuestsFu:AcceptQuest()
	if self.db.profile.rememberQuestgiver then
		if (not UnitName("npc")) or UnitIsPlayer("npc") then
			self.db.char.questgivers[GetTitleText()] = L["Item or other player"]
		else
			local x,y = GetPlayerMapPosition("player")
			self.db.char.questgivers[GetTitleText()] = L["%s in %s (%d,%d)"]:format(UnitName('npc'), GetZoneText(), floor(x*100), floor(y*100))
		end
	end
	return self.hooks["AcceptQuest"]()
end

function QuestsFu:Quest_Lost(name, id)
	if self.db.profile.useOwnTracker then
		self.db.char.watchedQuests[name] = nil
	end
	if self.questid == id then
		self.questid = nil
		self:UpdateQuestDetail()
	end
	if self.db.profile.rememberQuestgiver then
		self.db.char.questgivers[name] = nil
	end
end

local autowatched = {}
function QuestsFu:ZoneChange()
	if not (self.db.profile.autoQuestWatch.zone or self.db.profile.autoQuestWatch.subzone) then return end
	local zone = GetRealZoneText()
	local subzone = GetSubZoneText()
	--Get rid of autowatched quests
	for i, name in pairs(autowatched) do
		local questid = select(8, quixote:GetQuest(table.remove(autowatched, i)))
		RemoveQuestWatch(questid)
	end
	for _, questid in quixote:QuestsByZone(zone) do
		local name = quixote:GetQuestById(questid)
		-- If we're already watching it, don't autowatch.  (This way it won't be de-watched when we leave the zone.)
		if not self.db.char.watchedQuests[name] then
			if self.db.profile.autoQuestWatch.zone then
				table.insert(autowatched, name)
				AddQuestWatch(questid)
			elseif self.db.profile.autoQuestWatch.subzone and subzone ~= "" then
				local objective, description = quixote:GetQuestText(questid)
				if string.find(objective, subzone, 1, true) or string.find(description, subzone, 1, true) then
					table.insert(autowatched, name)
					AddQuestWatch(questid)
				end
			end
		end
	end
end

function QuestsFu:Quixote_Leaderboard_Update(name, id, lid, desc, numHad, numGot, numNeeded)
	if (not AUTO_QUEST_WATCH) and self.db.profile.autoQuestWatch.progress then
		AddQuestWatch(id)
	end
	--Announce to the party:
	if self.db.profile.partynotify.own and GetNumPartyMembers() > 0 then
		SendChatMessage(("%s - %s: %d/%d"):format(name, desc, numGot, numNeeded), "PARTY")
	end
end

function QuestsFu:Quixote_Quest_Gained(name, id)
	if self.db.profile.autoQuestWatch.gained then
		AddQuestWatch(id)
	end
	if self.db.profile.autoQuestWatch.zone then
		local zone = GetRealZoneText()
		local questzone = select(7, quixote:GetQuestById(id))
		if zone == questzone then
			AddQuestWatch(id)
		end
	elseif self.db.profile.autoQuestWatch.subzone then
		local subzone = GetSubZoneText()
		local objective, description = quixote:GetQuestText(id)
		if subzone ~= "" and (string.find(objective, subzone, 1, true) or string.find(description, subzone, 1, true)) then
			table.insert(autowatched, name)
			AddQuestWatch(id)
		end
	end
end

----------------------------------------------------
-- Sound on complete
----------------------------------------------------

function QuestsFu:Quixote_Quest_Complete(name, id)
	if self.db.profile.soundOnComplete then
		PlaySoundFile("Sound\\Creature\\Peon\\PeonBuildingComplete1.wav")
	end
end

----------------------------------------------------
-- Notifications
----------------------------------------------------

function QuestsFu:Quixote_Party_Leaderboard_Update(sender, title, description, numHad, numGot, numNeeded)
	message = ("(%s:%s) %s: %d/%d"):format(sender,title,description,numGot,numNeeded)
	if self.db.profile.partynotify.message then
		UIErrorsFrame:AddMessage(message, 1.0, 1.0, 0.0, 1.0)
	end
	if self.db.profile.partynotify.chat then
		self:Print(message)
	end
	
	self:UpdateTracker()
end

----------------------------------------------------
-- Quest tracker frame
----------------------------------------------------

function QuestsFu:UpdateTracker()
	if self.db.profile.useOwnTracker then
		if not tablet:IsRegistered('QuestsFu_Tracker') then
			tablet:Register('QuestsFu_Tracker', 'detachedData', self.db.profile.tracker_data, 'cantAttach', true, 'dontHook', true,
				'hideWhenEmpty', true, 'strata', self.db.profile.tracker_strata,
				'menu', function() dewdrop:FeedAceOptionsTable(optionsTable.args.tracker) end,
				'children', function()
					tablet:SetTitle("Tracker")
					local cat = tablet:AddCategory('columns', 2, 'child_func', 'QuestClick', 'child_arg1', self)
					for questName in pairs(self.db.char.watchedQuests) do self:AddQuestToCategory(cat, questName, false, true, true) end
				end)
			tablet:Open('QuestsFu_Tracker')
		end
		tablet:Refresh('QuestsFu_Tracker')
		QuestWatchFrame:Hide()
	end
end

----------------------------------------------------
-- Quest detail frame
----------------------------------------------------

local escapeDetector
function QuestsFu:UpdateQuestDetail()
	if self.db.profile.useOwnQuestDetail then
		if not tablet:IsRegistered('QuestsFu_Detail') then
			local qualitycolors = {
				[0] = "|cff9d9d9d",
				[1] = "|cffffffff",
				[2] = "|cff1eff00",
				[3] = "|cff0070dd",
				[4] = "|cffa335ee",
				[5] = "|cffff8000"
			}
			
			local questgiverPat = '^' .. L["%s in %s (%d,%d)"]:gsub("([%(%)%.%*%+%-%[%]%?%^%$%%])", "%%%1"):gsub("%%%%s", "(.*)"):gsub("%%%%d", "(%%d+)") .. '$'
			
			local showingDescription = true
			local function itemClick(link)
				if IsControlKeyDown() then
					DressUpItem(id)
				elseif IsShiftKeyDown() then
					if not ChatFrameEditBox:IsVisible() then
						ChatFrameEditBox:Show()
					end
					ChatFrameEditBox:Insert(link)
				else
					ShowUIPanel(ItemRefTooltip)
					if not ItemRefTooltip:IsVisible() then
						ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
					end
					ItemRefTooltip:SetHyperlink(link)
				end
			end
			local function hideQuestDetail()
				self.questid = nil
				tablet:Refresh('QuestsFu_Detail')
			end
			local function toggleDescription()
				showingDescription = not showingDescription
				tablet:Refresh('QuestsFu_Detail')
			end
			local function abandonQuestPopup(questid)
				local startingQuestLogSelection = GetQuestLogSelection()
				SelectQuestLogEntry(questid)
				SetAbandonQuest()
				local items = GetAbandonQuestItems()
				if items then
					StaticPopup_Hide("ABANDON_QUEST")
					StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
				else
					StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
					StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
				end
				SelectQuestLogEntry(startingQuestLogSelection)
			end
			local directions = L["bunch-of-directions"]
			local function colorizeQuestText(text, ...)
				--This is copied in approach from MonkeyQuestLog, note.
				for i = 1, select('#', ...) do
					local val = select(i, ...)
					if val then
						local val2 = val:gsub("([^%s])", "|cffffffff%1|r") -- color each word so it doesn't mess up later
						text = text:gsub(val, val2)
					end
				end
				
				text = text:gsub("([^%a%d%|])(%d+)([^%a%d%|])", "%1|cff00ff00%2|r%3") -- Match numbers
				text = text:gsub("([^%a%d%|])(%d+)([^%a%d%|])", "%1|cff00ff00%2|r%3") -- Match numbers again

				if ( GetLocale()=="koKR" ) then
					text = text:gsub("("..UnitName("player")..")", "|cffffffff%1|r") -- Match uppercase words. (Korean fix)
					text = text:gsub("(.+) ([^%s%|]+)에게 ", "%1 |cffffffff%2|r에게 ") -- Match uppercase words.
					text = text:gsub("(.+) ([^%s%|]+)(%|1으로;로;) ", "%1 |cffffffff%2|r%3 ") -- Run it twice or it'll skip words (TODO: figure out a better pattern to avoid this, if possible)
				else
					text = text:gsub("([%w,]+) (%u[%w'%-]+ %u[%w'%-]+)", "%1 |cffffffff%2|r") -- Match uppercase words.
					text = text:gsub("([%w,]+) (%u[%w'%-]+)", "%1 |cffffffff%2|r") -- Run it twice or it'll skip words (TODO: figure out a better pattern to avoid this, if possible)
				end
				
				for _, direction in ipairs(directions) do
					text = text:gsub(direction, "%1|cffffffff%2|r%3") -- Match direction
				end
				return text
			end
			
			escapeDetector = CreateFrame("Frame", "QuestsFu_escapeDetector")
			escapeDetector:SetScript("OnHide", hideQuestDetail)
			table.insert(UISpecialFrames, escapeDetector:GetName())
			
			tablet:Register('QuestsFu_Detail', 'detachedData', self.db.profile.detail_data,
				'dontHook', true, 'hideWhenEmpty', true,-- 'showTitleWhenDetached', true,
				'minWidth', 300, 'strata', self.db.profile.detail_strata, 'cantAttach', true,
				'menu', function() dewdrop:FeedAceOptionsTable(optionsTable.args.detail) end,
				'children', function()
					tablet:SetTitle("Detail")
					if self.questid then
						local startingQuestLogSelection = GetQuestLogSelection()
						SelectQuestLogEntry(self.questid)
						
						local title, level, tag, suggestedGroup, complete, nObjectives, zone, id = quixote:GetQuest(self.questid)
						local objective, description = quixote:GetQuestText(self.questid)
						
						--Title and objective:
						if not self.db.profile.color.title then title = crayon:White(title) end
						if not self.db.profile.color.level then level = crayon:White(questLevel) end
						local color = GetDifficultyColor(level)
						
						local giverName, giverZone
						if questgiverPat and self.db.char.questgivers[title] then
							giverName, giverZone = self.db.char.questgivers[title]:match(questgiverPat)
						end
						
						local cat = tablet:AddCategory('columns', 1, 'text', self:MakeTag(level, tag)..title,
							'textR', (color and color.r) or 1, 'textG', (color and color.g) or 1, 'textB', (color and color.b) or 1,
							'size', tablet:GetHeaderFontSize(), 'justify', 'CENTER',
							'func', self.QuestClick, 'arg1', self, 'arg2', id) --Do this instead of closing the tablet because it seems permanently blank on its second open otherwise.
						if complete then
							cat:AddLine('text', complete, 'textR', (complete == L["(failed)"]) and 1 or 0, 'textG', (complete == L["(done)"]) and 1 or 0, 'textB', 0, 'justify', 'CENTER')
						end
						if suggestedGroup and suggestedGroup > 0 then
							cat:AddLine('text', (QUEST_SUGGESTED_GROUP_NUM):format(suggestedGroup), 'justify', 'CENTER')
						end
						cat:AddLine('text', colorizeQuestText(objective, giverName, giverZone), 'wrap', true)
						
						--Leaderboard:
						if nObjectives > 0 then
							cat = tablet:AddCategory('columns', 2)
							for i=1,nObjectives do
								self:AddLeaderboardToCategory(cat, self.questid, i, tablet:GetNormalFontSize(), 0)
							end
						end
						
						--Given by?
						if self.db.profile.rememberQuestgiver then
							if self.db.char.questgivers[title] then
								cat = tablet:AddCategory('columns', 2, 'text', "Given by")
								cat:AddLine('text', colorizeQuestText(self.db.char.questgivers[title], giverName, giverZone), 'indentation', 2)
							end
						end
						
						--Requires money?
						if GetQuestLogRequiredMoney() > 0 then
							cat = tablet:AddCategory('columns', 2, 'text', "Required money: "..abacus:FormatMoneyFull(GetQuestLogRequiredMoney()), 'showWithoutChildren', true)
						end
						
						--Description:
						cat = tablet:AddCategory('columns', 1, 'text', "Description", 'func', toggleDescription, 'showWithoutChildren', true,
							'checked', true, 'hasCheck', true, 'checkIcon', (not showingDescription) and "Interface\\Buttons\\UI-PlusButton-Up" or "Interface\\Buttons\\UI-MinusButton-Up")
						if showingDescription then
							cat:AddLine('text', colorizeQuestText(description, giverName, giverZone), 'wrap', true)
						end
						
						--Rewards:
						cat = tablet:AddCategory('columns', 2, 'text', "Rewards")
						local numChoices = GetNumQuestLogChoices(self.questid) -- The ones you choose from
						local numRewards = GetNumQuestLogRewards(self.questid) -- The ones you always get
						local rewardMoney = GetQuestLogRewardMoney(self.questid) -- Moolah!
						local rewardSpellTexture, rewardSpellName = GetQuestLogRewardSpell() -- Sometimes there's a spell
						
						if numChoices > 0 or numRewards > 0 or rewardMoney > 0 then
							if numChoices > 0 then
								cat:AddLine('text', "Choose from:", 'indentation', 2)
								for i=1, numChoices, 1 do
									local name, texture, numItems, quality, isUsable = GetQuestLogChoiceInfo(i)
									cat:AddLine('text', (quality and qualitycolors[quality] or '')..'['..(name or '<not yet cached>')..']'..((numItems > 1) and (" x"..numItems) or ""), 'indentation', 4,
										'checked', true, 'hasCheck', true, 'checkIcon', texture,
										'func', itemClick, 'arg1', GetQuestLogItemLink('choice', i))
								end
							end
							if numRewards > 0 then
								cat:AddLine('text', "Always receive:", 'indentation', 2)
								for i=1, numRewards, 1 do
									local name, texture, numItems, quality, isUsable = GetQuestLogRewardInfo(i)
									cat:AddLine('text', (quality and qualitycolors[quality] or '')..'['..(name or '<not yet cached>')..']'..((numItems > 1) and (" x"..numItems) or ""), 'indentation', 4,
										'checked', true, 'hasCheck', true, 'checkIcon', texture,
										'func', itemClick, 'arg1', GetQuestLogItemLink('reward', i))
								end
							end
							if rewardSpellTexture then
								cat:AddLine('text', "Spell:", 'indentation', 2)
								cat:AddLine('text', rewardSpellName, 'indentation', 4,
									'checked', true, 'hasCheck', true, 'checkIcon', rewardSpellTexture)
							end
							if rewardMoney > 0 then
								cat:AddLine('text', 'Money: '..abacus:FormatMoneyFull(rewardMoney, true), 'indentation', 2)
							end
						else
							cat:AddLine('text', "Nothing.", 'indentation', 2)
						end
						
						cat = tablet:AddCategory('columns', 1, 'text', "Options")
						--Shareable:
						if GetQuestLogPushable() then
							cat:AddLine('text', "Share", 'func', 'ShareQuest', 'arg1', self, 'arg2', self.questid)
						end
						cat:AddLine('text', "Abandon", 'func', abandonQuestPopup, 'arg1', self.questid)
						cat:AddLine('text', "Close", 'func', hideQuestDetail, 'textR', 1, 'textG', 0, 'textB', 0, 'justify', 'RIGHT')
						
						SelectQuestLogEntry(startingQuestLogSelection)
					end
				end)
			tablet:Open('QuestsFu_Detail')
		end
		escapeDetector:Show()
		tablet:Refresh('QuestsFu_Detail')
	end
end

----------------------------------------------------
-- Utility functions
----------------------------------------------------

function QuestsFu:MakeTag(level, tag)
	return "[" .. level .. tag .. "] "
end

function QuestsFu:MembersOnQuest(questid)
	--Return the number of party members on a quest.  Returns zero if not in a party.
	local n = 0
	local party = GetNumPartyMembers()
	if party > 0 then
		for i = 1,party do
			n = n + (IsUnitOnQuest(questid, "party"..i) and 1 or 0)
		end
	end
	return n
end

function QuestsFu:ShowLog(questid)
	if self.db.profile.useOwnQuestDetail then
		if self.questid ~= questid then
			self.questid = questid
		else
			self.questid = nil
		end
		self:ScheduleEvent(self.UpdateQuestDetail, 0, self) --Schedule an event because tablet bugs out otherwise.
	else
		if QuestLogFrame:IsVisible() then
			if self.lastIndex == questid then
				HideUIPanel(QuestLogFrame)
			end
		else
			ShowUIPanel(QuestLogFrame)
		end
		local numEntries = GetNumQuestLogEntries()
		if (numEntries > QUESTS_DISPLAYED) then
			if (questid < numEntries - QUESTS_DISPLAYED) then
				FauxScrollFrame_SetOffset(QuestLogListScrollFrame, questid - 1)
				QuestLogListScrollFrameScrollBar:SetValue((questid - 1) * QUESTLOG_QUEST_HEIGHT)
			else
				FauxScrollFrame_SetOffset(QuestLogListScrollFrame, numEntries - QUESTS_DISPLAYED)
				QuestLogListScrollFrameScrollBar:SetValue((numEntries - QUESTS_DISPLAYED) * QUESTLOG_QUEST_HEIGHT)
			end
		end
		
		SelectQuestLogEntry(questid)
		QuestLog_SetSelection(questid)
		self.lastIndex = questid
	end
end

function QuestsFu:ShareQuest(questid)
	-- Share the quest with nearby party members.
	local wasSelected = GetQuestLogSelection()
	SelectQuestLogEntry(questid)
	if GetQuestLogPushable() and GetNumPartyMembers() > 0 then
		QuestLogPushQuest()
	end
	SelectQuestLogEntry(questid)
end
